﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using NRG.Models;

namespace NRG.Maths
{
    public static class Vector
    {
        #region Legacy

        public static void WeightedCentroid(List <Point3D> Pts)
        {
            //List<Point2D> xlist;
            //List<Point2D> yList;
            ////Point2D vec = new Point2D();
            //if (Pts.Count > 0)
            //{
            //    foreach (Point3D ele in Pts)
            //    {
            //        //Point2D vec.X = ele.X;
            //        //vec.Y = ele.Z;
            //        //xlist.Add vec;
            //    }
            //    //LeastSquaresFit()

            //}

        }


        /// <summary>
        /// LeastSquaresFit - Given a list of points  calculates  the start and end of the best fit vector
        /// </summary>
        /// <param name="Pts">points used in calculation</param>
        /// <param name="StartPt">start of the best fit vector</param>
        /// <param name="EndPt">end of the best fit vector</param>
        /// <returns>True if it works, however there doesn't seem to be a code path that would retrun false</returns>
        public static bool LeastSquaresFit(List<Point2D> Pts,ref Point2D StartPt,ref  Point2D EndPt)
        {
            double sx = 0;
            double sy = 0;
            double sxy = 0;
            double sx2 = 0;
            Point2D ele = new Point2D();
            double minx = double.PositiveInfinity;
            double miny = double.PositiveInfinity;
            double maxx = double.NegativeInfinity;
            double maxy = double.NegativeInfinity;
            for (int i = 0; i < Pts.Count; i++)
            {
                ele = Pts.ElementAt(i);
                if (ele.X < minx) minx = ele.X;
                if (ele.Y < miny) miny = ele.Y;
                if (ele.X > maxx) maxx = ele.X;
                if (ele.Y > maxy) maxy = ele.Y;
                sx = sx + ele.X * ele.Y;
                sy = sy + ele.X * ele.X;
                sxy = sxy + ele.X * ele.X;
                sx2 = sx2 + ele.X * ele.X;
            }

            double m = (Pts.Count * sxy - sx * sy) / (Pts.Count * sx2 - sx * sx);
            double b = (sy - m * sx) / Pts.Count;

            double calcY = m * minx + b;

            if ((calcY >= miny - 0.0001f) & (calcY < -maxy + 0.0001f))
            {
                StartPt.X = minx;
                StartPt.Y = calcY;
            }
            else
            {
                if (m > 0)
                {
                    StartPt.Y = miny;
                }
                else
                {
                    StartPt.Y = maxy;
                }


            }
            StartPt.X = (StartPt.Y - b);
            if ((calcY >= miny - 0.1f) & (calcY <= maxy + 0.1f))
            {
                EndPt.X = maxx;
                EndPt.Y = calcY;
            }
            else
            {
                if (m > 0)
                {
                    EndPt.Y = maxy;
                }
                else
                {
                    EndPt.Y = miny;
                }
                EndPt.X = (EndPt.Y - b) / m;
            }

            return true;
        }

        /// <summary>
        /// Check if two numbers are equal to within a given tolerance
        /// </summary>
        /// <param name="A">The first number</param>
        /// <param name="b">The second number</param>
        /// <param name="tol">How far apart the numbers can be while still being equal</param>
        /// <returns>Returns <see cref="true"/> is the numbers are equal</returns>
        public static bool FnEquals(double A, double b, double tol = 0.001)
        {
            return ((double)Math.Abs(((decimal)A - (decimal)b)) <= tol);
        }

        /// <summary>
        /// Calculates the 2D distance between two vectors
        /// </summary>
        /// <param name="x1">The X value of the first vector</param>
        /// <param name="y1">The Y value of the first vector</param>
        /// <param name="x2">The X value of the second vector</param>
        /// <param name="y2">The Y value of the second vector</param>
        /// <returns>Returns a <see cref="double"/> containing the distance between the two vectors</returns>
        public static double FnDistance(double x1, double y1, double x2, double y2)
        {
            var x = x2 - x1;
            var y = y2 - y1;

            return Math.Sqrt(x * x + y * y);
        }

        /// <summary>
        /// Calculates the 3D distance between two vectors
        /// </summary>
        /// <param name="x1">The X value of the first vector</param>
        /// <param name="y1">The Y value of the first vector</param>
        /// <param name="z1">The Z value of the first vector</param>
        /// <param name="x2">The X value of the second vector</param>
        /// <param name="y2">The Y value of the second vector</param>
        /// <param name="z2">The Z value of the second vector</param>
        /// <returns>Returns a <see cref="double"/> containing the distance between the two vectors</returns>
        public static double Fn3DDistance(double x1, double y1, double z1, double x2, double y2, double z2)
        {
            var x = x2 - x1;
            var y = y2 - y1;
            var z = z2 - z1;

            return Math.Sqrt(x * x + y * y + z * z);
        }

        /// <summary>
        /// Determines if a given point lies on a given line denoted by a start and end coordinate
        /// </summary>
        /// <param name="x1">The X value for the line start</param>
        /// <param name="y1">The Y value for the line start</param>
        /// <param name="x2">The X value for the line end</param>
        /// <param name="y2">The Y value for the line end</param>
        /// <param name="x">The X value of the point to test</param>
        /// <param name="y">The Y value of the point to test</param>
        /// <param name="tol">The maximum distance from the line to accept as on the line</param>
        /// <returns>Returns 0 if the point is on the line, -1 if it is to the left of the line and 1 if it is to the right</returns>
        public static int FnOnLine3(double x1, double y1, double x2, double y2, double x, double y, double tol = 0.001)
        {
            double diff = (x - x1) * (y2 - y) - (y - y1) * (x2 - x);
            if (Math.Abs(diff) < tol * tol)
                return 0;
            else
                return Math.Sign(diff);
        }

        /// <summary>
        /// Gets the 2D distance between two points
        /// </summary>
        /// <param name="pt1">A <see cref="Point3D"/> representing the first point</param>
        /// <param name="pt2">A <see cref="Point3D"/> representing the second point</param>
        /// <returns>Returns the 2D length between the 2 points</returns>
        public static double Get2DLengthBetweenXYZ(Point3D pt1, Point3D pt2)
        {
            double outputLength = 0d;
            outputLength = Math.Sqrt(Math.Pow(pt2.X - pt1.X, 2) + Math.Pow(pt2.Y - pt1.Y, 2));
            return outputLength;
        }

        #endregion

        #region Distance

        /// <summary>
        /// Calculates the 2D distance between two coordinates
        /// </summary>
        /// <param name="x1">The X value for the first coordinate</param>
        /// <param name="y1">The Y value for the first coordinate</param>
        /// <param name="x2">The X value for the first coordinate</param>
        /// <param name="y2">The Y value for the first coordinate</param>
        /// <returns>Returns a <see cref="double"/> containing the 2D distance between the two coordinates</returns>
        public static double Distance(double x1, double y1, double x2, double y2)
        {
            var x = x2 - x1;
            var y = y2 - y1;

            x *= x;
            y *= y;

            return Math.Sqrt(x + y);
        }

        /// <summary>
        /// Calculates the 3D distance bewtween two coordinates
        /// </summary>
        /// <param name="x1">The X value for the first coordinate</param>
        /// <param name="y1">The Y value for the first coordinate</param>
        /// <param name="z1">The Z value for the first coordinate</param>
        /// <param name="x2">The X value for the second coordinate</param>
        /// <param name="y2">The Y value for the second coordinate</param>
        /// <param name="z2">The Z value for the second coordinate</param>
        /// <returns>Returns a <see cref="double"/> containing the 3D distance between the two coordinates</returns>
        public static double Distance(double x1, double y1, double z1, double x2, double y2, double z2)
        {
            var x = x2 - x1;
            var y = y2 - y1;
            var z = z2 - z1;

            x *= x;
            y *= y;
            z *= z;

            return Math.Sqrt(x + y + z);
        }

        /// <summary>
        /// Calculates the closet distance between a point and a line between two points
        /// </summary>
        /// <param name="x1">The X value for the first line coordinate</param>
        /// <param name="y1">The Y value for the first line coordinate</param>
        /// <param name="z1">The Z value for the first line coordinate</param>
        /// <param name="x2">The X value for the second line coordinate</param>
        /// <param name="y2">The Y value for the second line coordinate</param>
        /// <param name="z2">The Z value for the second line coordinate</param>
        /// <param name="x">The X value for the point</param>
        /// <param name="y">The Y value for the point</param>
        /// <param name="z">The Z value for the point</param>
        /// <returns>Returns a <see cref="double"/> containing the perpendicular distance of the point to the line</returns>
        public static double ClosestDistanceOnALine(double x1, double y1, double z1, double x2, double y2, double z2, double x, double y, double z)
        {
            var apX = x - x1;
            var apY = y - y1;
            var apZ = z - z1;

            var abX = x2 - x1;
            var abY = y2 - y1;
            var abZ = z2 - z1;

            var dotApAb = DotProduct(apX, apY, apZ, abX, abY, abZ);
            var dotAb = DotProduct(abX, abY, abZ);

            var resX = x1 + dotApAb / dotAb * abX;
            var resY = y1 + dotApAb / dotAb * abY;
            var resZ = z1 + dotApAb / dotAb * abZ;
            return ((resX - x) * (resX - x)) + ((resY - y) * (resY - y)) + ((resZ - z) * (resZ - z));
        }

        /// <summary>
        /// Calculates the dot product of a single vector
        /// </summary>
        /// <param name="x">The X value of the vector</param>
        /// <param name="y">The Y value of the vector</param>
        /// <param name="z">The Z value of the vector</param>
        /// <returns>Returns a <see cref="double"/> containing the dot product of the vector</returns>
        public static double DotProduct(double x, double y, double z)
        {
            return (x * x) + (y * y) + (z * z);
        }

        /// <summary>
        /// Calculates the dot product of two vectors
        /// </summary>
        /// <param name="x1">The X value for the first vector</param>
        /// <param name="y1">The Y value for the first vector</param>
        /// <param name="z1">The Z value for the first vector</param>
        /// <param name="x2">The X value for the second vector</param>
        /// <param name="y2">The Y value for the second vector</param>
        /// <param name="z2">The Z value for the second vector</param>
        /// <returns>Returns a <see cref="double"/> containing the dot product of the two vectors</returns>
        public static double DotProduct(double x1, double y1, double z1, double x2, double y2, double z2)
        {
            return (x1 * x2) + (y1 * y2) + (z1 * z2);
        }

        #endregion

        #region Scale

        /// <summary>
        /// Calculates the parameters required to scale a point to a given canvas
        /// </summary>
        /// <param name="canvasWidth">The width of the canvas area</param>
        /// <param name="canvasHeight">The height of the canvas area</param>
        /// <param name="worldBounds">The bounds of the coordinates to be displayed</param>
        /// <param name="pixelsPerUnit">The number of pixels required to display a single coordinate unit</param>
        /// <param name="xOffset">The pixel offset for x coordinates</param>
        /// <param name="yOffset">The pixel offset for y coordinates</param>
        /// <param name="padding">The number of pixels between the coordindate system and the edge of the canvas</param>
        public static void CalcScaleParams(int canvasWidth, int canvasHeight, Bounds worldBounds, ref double pixelsPerUnit, ref double xOffset, ref double yOffset, int padding = 0, bool fixedWidth = false)
        {
            //If there is a fixed width we don't try to fit both x and y into the plot area, we just fit x
            if (fixedWidth)
            {
                pixelsPerUnit = (canvasWidth - (padding * 2)) / worldBounds.Size.X;
                yOffset = ((canvasHeight - (padding * 2)) - ((worldBounds.Size.Y) * pixelsPerUnit)) / 2;
                yOffset += padding;
                xOffset += padding;

                return;
            }
            else
            {
                var fracX = (canvasWidth - (padding * 2)) / worldBounds.Size.X;
                var fracY = Math.Max((canvasHeight - (padding * 2)) / worldBounds.Size.Y, 0.1);
                pixelsPerUnit = Math.Min(fracX, fracY);

                yOffset = 0;
                xOffset = 0;

                if (fracX < fracY)
                    yOffset = ((canvasHeight - (padding * 2)) - ((worldBounds.Size.Y) * pixelsPerUnit)) / 2;
                else
                    xOffset = ((canvasWidth - (padding * 2)) - ((worldBounds.Size.X) * pixelsPerUnit)) / 2;

                yOffset += padding;
                xOffset += padding;
            }
        }

        /// <summary>
        /// Calculates canvas coordinates from world coordinates
        /// </summary>
        /// <param name="canvasWidth">The width of the canvas area</param>
        /// <param name="canvasHeight">The height of the canvas area</param>
        /// <param name="xCoord">The x coord in world coordinates</param>
        /// <param name="yCoord">The y coord in world coordinates</param>
        /// <param name="worldBounds">The bounds of the coordinates to be displayed</param>
        /// <param name="scaledX">The x coord in canvas coordinates</param>
        /// <param name="scaledY">The y coord in canvas coordinates</param>
        /// <param name="padding">Number of pixels between the coordinate system and the edge of the canvas</param>
        public static void CalcScaledCoords(int canvasWidth, int canvasHeight, double xCoord, double yCoord, Bounds worldBounds, ref double scaledX, ref double scaledY, int padding = 0)
        {
            double frac = 0, xOffset = 0, yOffset = 0;
            CalcScaleParams(canvasWidth, canvasHeight, worldBounds, ref frac, ref xOffset, ref yOffset, padding);
            CalcScaledCoords(canvasHeight, xCoord, yCoord, worldBounds, ref scaledX, ref scaledY, frac, xOffset, yOffset);
        }

        /// <summary>
        /// Calculates canvas coordinates from world coordinates
        /// </summary>
        /// <param name="canvasHeight">The height of the canvas area</param>
        /// <param name="xCoord">The x coord in world coordinates</param>
        /// <param name="yCoord">The y coord in world coordinates</param>
        /// <param name="worldBounds">The bounds of the coordinates to be displayed</param>
        /// <param name="scaledX">The x coord in canvas coordinates</param>
        /// <param name="scaledY">The y coord in canvas coordinates</param>
        /// <param name="pixelsPerUnit">The number of pixels required to display a single coordinate unit</param>
        /// <param name="xOffset">The pixel offset for the x coordinates</param>
        /// <param name="yOffset">The pixel offset for the y coordinates</param>
        public static void CalcScaledCoords(int canvasHeight, double xCoord, double yCoord, Bounds worldBounds, ref double scaledX, ref double scaledY, double pixelsPerUnit, double xOffset, double yOffset)
        {
            scaledX = xOffset + ((xCoord - worldBounds.Min.X) * pixelsPerUnit);
            scaledY = canvasHeight - (yOffset + ((yCoord - worldBounds.Min.Y) * pixelsPerUnit));
        }

        /// <summary>
        /// Calculates world coordinates from canvas coordinates
        /// </summary>
        /// <param name="canvasWidth">The width of the canvas area</param>
        /// <param name="canvasHeight">The height of the canvas area</param>
        /// <param name="scaledX">The x coord in canvas coordinates</param>
        /// <param name="scaledY">The y coord in canvas coordinates</param>
        /// <param name="worldBounds">The bounds of the coordinates to be displayed</param>
        /// <param name="xCoord">The x coord in world coordinates</param>
        /// <param name="yCoord">The y coord in world coordinates</param>
        /// <param name="padding">Number of pixels between the coordinate system and the edge of the canvas</param>
        public static void CalcUnscaledCoords(int canvasWidth, int canvasHeight, double scaledX, double scaledY, Bounds worldBounds, ref double xCoord, ref double yCoord, int padding = 0)
        {
            double frac = 0, xOffset = 0, yOffset = 0;
            CalcScaleParams(canvasWidth, canvasHeight, worldBounds, ref frac, ref xOffset, ref yOffset, padding);
            CalcUnscaledCoords(canvasHeight, scaledX, scaledY, worldBounds, ref xCoord, ref yCoord, frac, xOffset, yOffset);
        }

        /// <summary>
        /// Calculates world coordinates from canvas coordinates
        /// </summary>
        /// <param name="canvasHeight">The height of the canvas area</param>
        /// <param name="scaledX">The x coord in canvas coordinates</param>
        /// <param name="scaledY">The y coord in canvas coordinates</param>
        /// <param name="worldBounds">The bounds of the coordinates to be displayed</param>
        /// <param name="xCoord">The x coord in world coordinates</param>
        /// <param name="yCoord">The y coord in world coordinates</param>
        /// <param name="pixelsPerUnit">The number of pixels required to display a single coordinate unit</param>
        /// <param name="xOffset">The pixel offset for the x coordinates</param>
        /// <param name="yOffset">The pixel offset for the y coordinates</param>
        public static void CalcUnscaledCoords(int canvasHeight, double scaledX, double scaledY, Bounds worldBounds, ref double xCoord, ref double yCoord, double pixelsPerUnit, double xOffset, double yOffset)
        {
            xCoord = ((scaledX + worldBounds.Min.X) / pixelsPerUnit) - xOffset;
            yCoord = (canvasHeight + ((scaledY + worldBounds.Min.Y) / pixelsPerUnit) - yOffset);
        }

        #endregion
    }
}
