﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using NRG.Services;

namespace NRG.Models
{
    public class Features
    {
        public class LineFeature : Feature
        {
            #region Setup

            public LineFeature()
            {

            }

            public LineFeature(string code, string[] args, bool isBreakLine = false)
            {
                this.Code = code.Replace("?", "").Replace("*", "");
                long col = 0;
                if (!long.TryParse(args[1], out col))
                {
                    R = 0;
                    G = 0;
                    B = 0;
                }
                else
                {
                    int r = 0, g = 0, b = 0;
                    Conversions.LongIntToRGB(col, ref r, ref g, ref b);

                    R = r;
                    G = g;
                    B = b;
                }

                double thickness = 0;
                string thicknessArgs = "";
                if (isBreakLine)
                    thicknessArgs = args[14];
                else
                    thicknessArgs = args[13];

                double.TryParse(thicknessArgs, out thickness);
                Thickness = thickness;
            }

            #endregion

            #region Methods

            public override Feature Copy()
            {
                return new LineFeature()
                {
                    Code = this.Code,
                    Thickness = this.Thickness,
                    R = this.R,
                    G = this.G,
                    B = this.B
                };
            }

            #endregion
        }

        public class PointFeature : Feature
        {
            #region Setup

            public PointFeature()
            {

            }

            public PointFeature(string code, string[] args)
            {
                this.Code = code.Replace("?", "").Replace("*", "");
                long col = 0;
                if (!long.TryParse(args[1], out col))
                {
                    R = 0;
                    G = 0;
                    B = 0;
                }
                else
                {
                    int r = 0, g = 0, b = 0;
                    Conversions.LongIntToRGB(col, ref r, ref g, ref b);
                    R = r;
                    G = g;
                    B = b;
                }

                double thickness = 0.2;
                if (!double.TryParse(args[11], out thickness))
                    thickness = 1;

                Thickness = thickness;
            }

            #endregion

            #region Methods

            public override Feature Copy()
            {
                return new PointFeature()
                {
                    Code = this.Code,
                    Thickness = this.Thickness,
                    R = this.R,
                    G = this.G,
                    B = this.B
                };
            }

            #endregion
        }
    }

    public class Feature
    {
        #region Properties

        public string Code { get; set; }
        public double Thickness { get; set; }
        public int R { get; set; }
        public int G { get; set; }
        public int B { get; set; }

        #endregion

        #region Methods

        public override string ToString()
        {
            return Code;
        }

        public virtual Feature Copy()
        {
            return null;
        }

        #endregion
    }
}
