﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using NRG.Services;
using NRG.Models;

namespace NRG.Volumes


{
    /// <summary>
    /// 
    /// </summary>
    class VolumeVector
    {
        public double X { get; set; }
        public double Y { get; set; }
        public double Z { get; set; } //Represents finished ground level
        public double ZB { get; set;} //Represents original ground level
    }

    public class VolumeResult
    {
        public double Cut { get; set; }
        public double Fill { get; set; }
    }


    class VolumeTriangle
        /* CCW triangle
                  v3
                  /\
                 /  \
                /    \  
             v1/______\v2 
             
        */
   
    {
        public VolumeVector Vector1 { get; set; }
        public VolumeVector Vector2 { get; set; }
        public VolumeVector Vector3 { get; set; }

        public VolumeResult GetCutAndFill()
        {

            double totalVolume = 0;
            VolumeResult v = new VolumeResult();


            totalVolume = Maths.Triangle.CalcTriPlanArea(Vector1.X,Vector1.Y, Vector2.X,Vector2.Y, Vector3.X, Vector3.Y) * (((Vector1.Z-Vector1.ZB)+ (Vector2.Z - Vector2.ZB) + (Vector3.Z - Vector3.ZB)) /3);

            //==========================================================================================================
            //not so special cases for when a triangle is totally in cut or totally in fill...

            if (Vector1.Z - Vector1.ZB >= 0 & Vector2.Z - Vector2.ZB >= 0 & Vector3.Z - Vector3.ZB >= 0) //All fill
            {
                v.Fill = totalVolume;
                return v;
            }

            else if (Vector1.Z - Vector1.ZB <= 0 & Vector2.Z - Vector2.ZB <= 0 & Vector3.Z - Vector3.ZB <= 0) //All cut
            {
                v.Fill = totalVolume;
                return v;
            }
            //==========================================================================================================

            //OK, I knew we wouldn't get that lucky! The triangle is in both cut and fill. I'm going to wing this for the time
            //being, without refering to the VB code - 'cos that's the way I roll.
            //We use vector intersections to work out which proportion is in cut and fill.
            //Unfortunately this is going to get a shit load more complicated...but for the time being:


            double diff1 = Vector1.Z - Vector1.ZB;
            double diff2 = Vector2.Z - Vector2.ZB;
            double diff3 = Vector3.Z - Vector3.ZB;

            Point2D p1 = new Point2D();
            Point2D p2 = new Point2D();
            Point2D p3 = new Point2D();
            bool isFill = false;
            double avgDepth = 0;


            //We need the odd one out...

            if ((diff1 >= 0 & diff2 >= 0)| (diff1 <= 0 & diff2 <= 0))
            {
                //vector  3  is the odd one out
                p1 = NRG.Maths.Interpolation.InterpolateOntoLine(Vector3.X, Vector3.Y, diff3, Vector1.X, Vector1.Y, diff1,0.0f);
                p3 = NRG.Maths.Interpolation.InterpolateOntoLine(Vector3.X, Vector3.Y, diff3, Vector2.X, Vector2.Y, diff2, 0.0f);
                p2.X = Vector3.X;
                p2.Y = Vector3.Y;
                isFill = diff1 >= 0 & diff2 >= 0;
                avgDepth = diff3 / 3;
            }
            else if ((diff2 >=0 & diff3 >=0) | (diff2 <= 0 & diff3 <= 0))
            {
                //vector 1 is the odd one out
                p1 = NRG.Maths.Interpolation.InterpolateOntoLine(Vector1.X, Vector1.Y, diff1, Vector3.X, Vector3.Y, diff3, 0.0f);
                p3 = NRG.Maths.Interpolation.InterpolateOntoLine(Vector1.X, Vector1.Y, diff1, Vector2.X, Vector2.Y, diff2, 0.0f);
                p2.X = Vector1.X;
                p2.Y = Vector1.Y;
                isFill = diff2 >= 0 & diff3 >= 0;
                avgDepth = diff1 / 3;
            }
            else if ((diff1 >=0 & diff3 >=0) | (diff1 <= 0 & diff3 <= 0))
            {
                //vector 2 is the odd one out
                p1 = NRG.Maths.Interpolation.InterpolateOntoLine(Vector2.X, Vector2.Y, diff2, Vector1.X, Vector1.Y, diff1, 0.0f);
                p3 = NRG.Maths.Interpolation.InterpolateOntoLine(Vector2.X, Vector2.Y, diff2, Vector3.X, Vector3.Y, diff3, 0.0f);
                p2.X = Vector2.X;
                p2.Y = Vector2.Y;
                isFill = diff1 >= 0 & diff3 >= 0;
                avgDepth = diff2 / 3;
            }

            double partialVol = Maths.Triangle.CalcTriPlanArea(p1.X,p1.Y,p2.X,p2.Y,p3.X,p3.Y) * avgDepth;

            if (isFill)
            {
                v.Fill = partialVol;
                v.Cut = totalVolume - partialVol;

            }
            else
            {
                v.Cut = partialVol;
                v.Fill= totalVolume - partialVol;

            }
            return v;
        }

        


    }

    
}
